//Database class
//Copyright:J.G. van den Berg
//          M. Zeldenrust N. laan 47
//          3066 GR Rotterdam
//          Nederland (Holland)
//          Email: jgvdberg@worldaccess.nl
//          Compuserve: 106200,1766



//Find command equates
#define FIND_PREV						1
#define FIND_NEXT						2
#define FIND_FIRST					3
#define FIND_LAST						4
#define FIND_EQ							5
#define FIND_GREQ						6
#define FIND_GR							7
#define FIND_LSEQ						8
#define FIND_LS							9

//Default number off cachepages
#define CACHEPAGES					40

//Error number list, see resources in DLL for description
#define ERR_FILE_SEEK			-1
#define ERR_FILE_CREATE		-2
#define ERR_FILE_NOTFOUND	-3
#define ERR_FILE_MODE			-4
#define ERR_FILE_DBOPEN		-5
#define ERR_FILE_PAGESIZE	-6
#define ERR_FILE_NOMEM		-7
#define ERR_FILE_NODB			-8
#define ERR_NAME_THERE		-9
#define ERR_FILE_TNF			-10
#define ERR_FILE_KNF			-11
#define ERR_FILE_FNF			-12
#define ERR_FILE_FLDTHERE	-13
#define ERR_FILE_WRITE		-14
#define ERR_FILE_READ			-15
#define ERR_FILE_CRMODE		-16
#define ERR_FILE_SIZE			-17
#define ERR_FILE_LOCK			-18
#define ERR_NO_TABLES			-20
#define ERR_NO_FIELDS			-21
#define ERR_FILE_INVLINK	-22
#define ERR_NO_KEYFIELDS	-23
#define ERR_FILE_NODBS		-24
#define ERR_FILE_RECPAGE	-25
#define ERR_FILE_KEYTHERE	-26
#define ERR_FILE_RECNF		-27
#define ERR_FILE_CHANGED	-28
#define ERR_FILE_KEYNF		-29
#define ERR_FILE_NOKEY		-30
#define ERR_FILE_FINDOPD	-31
#define ERR_FILE_EOIX			-32
#define ERR_FILE_NOREC		-33
#define ERR_FILE_PRIONE		-34
#define ERR_FILE_PRIDUP		-35
#define ERR_FILE_PKCHGD		-36
#define ERR_FILE_QNF			-37
#define ERR_FILE_LOCKED		-38

//Default page size
#define DEFAULTPAGESIZE		4096

//Field Type equates
#define FIELD_BYTE				1
#define FIELD_INT					2
#define FIELD_SHORT				3
#define FIELD_BOOL				4
#define FIELD_WORD				5
#define FIELD_LONG				6
#define FIELD_DWORD				7
#define FIELD_FLOAT				8
#define FIELD_DOUBLE			9
#define FIELD_CHAR				10
#define FIELD_MEMO				11

//Sort modes
#define SORT_ASC					1
#define SORT_DESC					2
#define SORT_FIND					3

//Database open modes
#define OPEN_NORMAL				1
#define OPEN_READONLY			2
#define OPEN_EXCLUSIVE		3

//Struct for memofield, add this struct to your record struct if you want a memofield in your record.
typedef struct tagMemoField
{
	long				Page;
	DWORD				Size;
} MEMOFLD, *LPMEMOFLD;


class CTvdbcDatabase
{
//Declares
public:
	CTvdbcDatabase();
	~CTvdbcDatabase();
	short			OpenDatabase(LPSTR FileName, WORD OpenType, LPSTR newXorMask = NULL);
	//Opens a database
	//Returns: short, error number if < 0, = 0 open ok
	//Parameters:
	//	Filename
	//		Name of file to open
	//	OpenType
	//		How to open file
	//			OPEN_NORMAL				Share, read, write
	//			OPEN_READONLY			Share, read
	//			OPEN_EXCLUSIVE		read, write
	//				In this mode your database is corrupt if your PC crashes, but adding
	//				and updating is very fast. In the other two modes your database is
	//				still ok after a PC crash.
	//	newXorMask
	//		Mask to de/encrypt the data in the database.

	short			CreateDatabase(LPSTR FileName, LPSTR newXorMask = NULL, WORD PageSize = DEFAULTPAGESIZE);
	//Creates a new database
	//Returns: short, error number if < 0, = 0 create ok
	//Parameters:
	//	Filename
	//		Name of file to create
	//	newXorMask
	//		Mask to de/encrypt the data in the database. You cannot open the database
	//		without this mask.
	//	PageSize
	//		Size of a record/key page in bytes. The PageSize must be a multiple of
	//		512 bytes. (PageSize - 12) must be greater than the size of your largest
	//		times two.
	//
	//After a CreateDatabase operation is only possible to use these commands:
	//	EndCreateDatabase*
	//	CreateTable*
	//	CreateTableField*
	//	CreateTableKey*
	//	CreateTableKeyField*
	//	GetErrorText
	//	AbortFatal
	//* These members cannot be called after a OpenDatabase command

	short			EndCreateDatabase();
	//Writes the create info to the file and closes the database
	//Returns: short, error number if < 0, = 0 end create ok
	//Parameters: none
	//Call this member after ypu created the tables, fields and keys.
	//After a CreateDatabase operation you cannot call CloseDatabase, you will
	//have to call this member.
	//Abort your app if this commands fails, there will be memory leaks.
	
	short			CloseDatabase();
	//Closes the database
	//Returns: short, error number if < 0, = 0 close ok

	short			CreateTable(LPSTR Name);
	//Creates a new table
	//Returns: TablesId (short) if > 0, error number if < 0
	//Parameters:
	//	Name
	//		Name of the table
	//
	//Starts the description of a struct. Always put these fields at the start of
	//your struct:
	//	DWORD		RecordId
	//	WORD		UpdateCounter
	//Never update the value of these fields, RecordId is incremented every time
	//you add an record. UpdateCounter is set to zero when you add an record, and
	//is incremented every time you update the record. RecordId is used to uniquely
	//identify the records, UpdateCounter is used to check if an other user has
	//changed the record between your find record and update/delete commands.
	//One key with recordid is added to your table, this key gets KeyId 1. This key
	//will be the primary key if you don't supply another primary key. Do not
	//add keyfields to this key.
	//The first table gets TableId 1, the second table gets TableId 2, etc..

	short			CreateTableField(short TableId, LPSTR FieldName, BYTE Type, WORD Size = 0);
	//Creates a field in a table
	//Returns: FieldId (short) if > 0, error number if < 0
	//Parameters:
	//	TableId
	//		Id of the table, see CreateTable
	//	FieldName
	//		Name of the tablefield
	//	Type
	//		Type of field, see Field Type equates
	//	Size
	//		Size of field, use only if Type = FIELD_CHAR
	//The first field in a table gets FieldId 1, the second field gets FieldId 2, etc..
	//FieldId 1 = RecordId and FieldId 2 = UpdateCounter, see CreateTable

	short			CreateTableKey(short TableId, LPSTR KeyName, BOOL Primary, BOOL Duplicates, short TableLinkId = 0);
	//Creates a key in a table
	//Returns: KeyId (short) if > 0, error number if < 0
	//Parameters:
	//	TableId
	//		Id of the table, see CreateTable
	//	KeyName
	//		Name of the tablekey
	//	Primary
	//		A primary key is used to link tables. A primary key must be 
	//		unique (Duplicates = FALSE)
	//	Duplicates
	//		Set to TRUE if the key can contain duplicates
	//	TableLinkId
	//		TableId of the Table this key is linked to. This key must at least 
	//		contain the same fields in the same order and the same size as the
	//		primary key of the linktable.
	//
	//The first key in a table gets KeyId 1, the second key gets KeyId 2, etc..
	//KeyId 1 = RecordId, see CreateTable
	
	short			CreateTableKeyField(short TableId, short KeyId, short FieldId, BYTE Sort = SORT_ASC, WORD Size = 0);
	//Creates a key in a table
	//Returns: 0 if ok, error number if < 0
	//Parameters:
	//	TableId
	//		Id of the table, see CreateTable
	//	KeyId
	//		Id of the key, see CreateTableKey
	//	FieldId
	//		Id of the field to add to the key, see CreateTableField
	//	Sort 
	//		How to sort the field, Ascending = SORT_ASC, Descending = SORT_DESC
	//	Size
	//		Size to use in sort, use only if FieldType = FIELD_CHAR
	
	void			SetCacheSize(short Pages = CACHEPAGES);
	//Creates a key in a table
	//Parameters:
	//	Pages
	//		Number of cache pages to create
	//
	//The Db Dll sets the optimum number of cache pages for the db. Use this 
	//member if you have a lot of links and experience performance problems.
	//The Db Dll cannot determine the optimum number of cache pages of a
	//db with a lot of links

	void			GetErrorText(short ErrorNo, LPSTR ErrorText);
	//Gets a errortext of an error returned by a member function.
	//Parameters:
	//	ErrorNo
	//		Error number returned by member function
	//	ErrorText
	//		Buffer to recieve errortext, buffer must at least 265 bytes
	//
	//Reads the errortext in the resource table, you can edit teh resource table.

	void			AbortFatal(short ErrorNo);
	//Aborts YOUR application and displays an error
	//Parameters:
	//	ErrorNo
	//		Error number returned by member function

	short			FindTableId(LPSTR TableName);
	//Returns the Id of a table
	//Returns: TableId, 0 if error
	//Parameters:
	//	TableName
	//		Name of the table

	short			FindFieldId(short TableId, LPSTR FieldName);
	//Returns the Id of a field
	//Returns: FieldId, 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldName
	//		Name of the field

	short			FindKeyId(short TableId, LPSTR KeyName);
	//Returns the Id of a key
	//Returns: KeyId, 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyName
	//		Name of the key

	short			AddRecord(short TableId, BOOL FailOnLinks = TRUE);
	//Adds a record to a table
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FailOnLinks
	//		Return an error if a link does not exist, zero values of a link are not
	//		checked.
	//
	//Use WriteTableRecord to write the record to the recordbuffer

	short			DeleteRecord(short TableId, BOOL FailOnLinks = TRUE, BOOL DeleteLinks = TRUE);
	//Delete a record from a table
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FailOnLinks
	//		Return an error if a link does not exist, zero values of a link are not
	//		checked.
	//	DeleteLinks
	//		Deletes the records in the linked tables if TRUE and FailOnLinks is TRIE
	//
	//Use FindRecord, QueryFindRecord or QueryGotoRecord to set the record to delete

	short			UpdateRecord(short TableId, BOOL FailOnLinks = TRUE);
	//Updates a record in a table
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FailOnLinks
	//		Return an error if a link does not exist, zero values of a link are not
	//		checked.
	//
	//Use WriteTableRecord to write the record to the recordbuffer
	//Use FindRecord, QueryFindRecord or QueryGotoRecord to set the record to update

	short			SetKey(short TableId, short KeyId);
	//Set the key for the FindRecord command
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyId
	//		Id of the key
	//
	//Use the WriteTableKeyField command to set the information to find

	short			FindRecord(short TableId, short FindOpd, BOOL ReadRecord = TRUE);
	//Find a record of a table
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FindOpd
	//		Way to find the record, see Find command equates
	//	ReadRecord
	//		Read the record, only read the record if you need it. You can use
	//		ReadTableKey and ReadTableKeyField to read the information returned in
	//		the key.
	//
	//Use ReadTableRecord to read the record out of the recordbuffer

	BOOL			CloseQuery(short QueryId);
	//Closes a query
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query

	short			CreateQuery(short ParentId, short TableId, short KeyId, short MinFindOpd = FIND_FIRST, short MaxFindOpd = FIND_LAST, BOOL DoLinks = FALSE, BOOL (*AcceptRecord)(short) = NULL);
	//Opens a new query
	//Returns QueryId > 0 if ok, < 0 if error
	//Parameters:
	//	ParentId
	//		QueryId of query to link this query to, this query will be called for 
	//		every record of the parent query. Only one record will be read.
	//	TableId
	//		Id of table to perform query on.
	//	KeyId
	//		Id of key to use
	//	MinFindOpd
	//		Find command for first record
	//	MaxFindOpd
	//		Find command for last record
	//	DoLinks
	//		Read the records of the linked tables
	//	AcceptRecord
	//		Callback, is called every a time a record is read, return TRUE to accept
	//		the record. This function called after the child querys are evalueted and
	//		the linked records have been read.


	BOOL			QuerySetCopyField(short QueryId, short No, short TableId, short FieldId);
	//Set the tablefield to copy into the keyfield.
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query
	//	No
	//		Number (0 - (No of Keyfields - 1)) of the keyfield to copy the 
	//		fieldvalue into
	//	TableId
	//		Id of table that contains the field
	//	FieldId
	//		Id of field to copy into to keyfield
	//
	//Every time a query executes its child querys the min and max key values will
	//be filled in. This member enables you to use dynamic key expressions

	BOOL			QuerySetMinKey(short QueryId, short No, LPVOID ptr);
	//Sets value for the minimum value of the key.
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query
	//	No
	//		Number (0 - (No of Keyfields - 1)) of the keyfield to copy the 
	//		ptr into
	//	ptr
	//		pointer to a buffer that contains the info to copy

	BOOL			QuerySetMaxKey(short QueryId, short No, LPVOID ptr);
	//Sets value for the maximum value of the key.
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query
	//	No
	//		Number (0 - (No of Keyfields - 1)) of the keyfield to copy the 
	//		ptr into
	//	ptr
	//		pointer to a buffer that contains the info to copy

	BOOL			QuerySetSeekKey(short QueryId, short No, LPVOID ptr);
	//Sets value for a seek operation on the query key.
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query
	//	No
	//		Number (0 - (No of Keyfields - 1)) of the keyfield to copy the 
	//		ptr into
	//	ptr
	//		pointer to a buffer that contains the info to copy
	//
	//Use this member if you want to perform a key value search on the query.
	//A search operation is performed if you use QueryFindRecord with 
	//(FindOpd = FIND_EQ or FIND_GREQ or FIND_GR or FIND_LSEQ or FIND_LS)

	short			QueryFindRecord(short QueryId, short FindOpd);
	//Execute the query
	//Returns FALSE if ok, < 0 if error
	//	QueryId
	//		Id of the Query
	//	FindOpd
	//		Way to find the record, see Find command equates
	//
	// FindOpd = FIND_FIRST
	//		Find the first record of the query
	// FindOpd = FIND_LAST
	//		Find the last record of the query
	// FindOpd = FIND_NEXT
	//		Find the next record of the query
	// FindOpd = FIND_PREV
	//		Find the prev record of the query

	BOOL			QueryGotoRecord(short QueryId, DWORD RecordId);
	//Jump to a query record number (not the RecordId of a table)
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	QueryId
	//		Id of the Query
	//	RecordId
	//		No of the record, 0 = first record of the query, 
	//		GetMaxQueryRecord() - 1 is the last record

	short			QueryGetSearchMode(short QueryId);
	//Returns current search mode of the query
	//Parameters:
	//	QueryId
	//		Id of the Query
	//
	//A query has three possible modes:
	//SORT_ASC
	//	The search is performed from the first record in the query
	//	GetCurrentQueryRecord(on first record) = 0
	//	GetCurrentQueryRecord(on last record) = ?
	//SORT_DESC
	//	The search is performed from the last record in the query
	//	GetCurrentQueryRecord(on first record) = ?
	//	GetCurrentQueryRecord(on last record) = GetMaxQueryRecord() - 1
	//SORT_FIND
	//	The search is performed from a searched record in the query
	//	GetCurrentQueryRecord(on first record) = ?
	//	GetCurrentQueryRecord(on last record) = ?

	short			GotoRecord(short TableId, DWORD RecordId);
	//Jumps to a record in a table
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	RecordId
	//		Id of record, See create table
	
	DWORD			GetCurrentQueryRecord(short QueryId);
	//Returns the current record of the query (not the RecordId of a table)
	//Parameters:
	//	QueryId
	//		Id of the Query

	DWORD			GetMaxQueryRecord(short QueryId);
	//Returns the maximum record of the query (not the RecordId of a table)
	//Parameters:
	//	QueryId
	//		Id of the Query
	//
	//The Db Dll uses the record count of the table times two for the expected count

	short			DupHeader(LPSTR FileName, WORD PageSize = 0);
	//Create a duplicate empty database, can be used for recover operations
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	FileName
	//		Name of the file to create
	//	PageSize
	//		New pagesize, = 0 use current pagesize
	//
	//Abort your app if this fails, memory leaks

	short			Recover(LPSTR FileName, WORD PageSize = 0);
	//Recovers a damaged database
	//Returns: 0 if ok, < 0 if error
	//Parameters:
	//	FileName
	//		Name of the databasefile to recover
	//	PageSize
	//		Pagesize of damaged database, = 0 use current pagesize
	//
	//How to recover:
	//1. The header is corrupt (OpenDatabase returns an error)
	//	Create an exact copy of the db (build a createdatabase in your app)
	//	Use recover on this copy
	//2. A page is damaged (the Db Dll aborts with an error)
	//	Open the database
	//	Use DupHeader
	//	Close the database
	//	Open the copy
	//	Use recover on this copy

	BOOL			ReadTableRecord(short TableId, LPVOID pRecord);
	//Read the record in the record buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	pRecord
	//		Pointer to your record struct

	BOOL			WriteTableRecord(short TableId, LPVOID pRecord);
	//Write your record to the record buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	pRecord
	//		Pointer to your record struct

	BOOL			ReadTableField(short TableId, short FieldId, LPVOID pRecord);
	//Read a field in the record buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldId
	//		Id of the field
	//	pRecord
	//		Pointer to your field

	BOOL			WriteTableField(short TableId, short FieldId, LPVOID pRecord);
	//Write a field to the record buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldId
	//		Id of the field
	//	pRecord
	//		Pointer to your field

	BOOL			ReadTableKey(short TableId, short KeyId, LPVOID pRecord);
	//Read a key in the key buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyId
	//		Id of the key
	//	pRecord
	//		Pointer to your key struct

	BOOL			WriteTableKey(short TableId, short KeyId, LPVOID pRecord);
	//Write a key to the key buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyId
	//		Id of the key
	//	pRecord
	//		Pointer to your key struct

	BOOL			ReadTableKeyField(short TableId, short KeyId, short No, LPVOID pRecord);
	//Read a keyfield in the key buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyId
	//		Id of the key
	//	No
	//		Number of the keyfield
	//	pRecord
	//		Pointer to your keyfield

	BOOL			WriteTableKeyField(short TableId, short KeyId, short No, LPVOID pRecord);
	//Write a keyfield to the key buffer
	//Returns TRUE if ok, FALSE if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	KeyId
	//		Id of the key
	//	No
	//		Number of the keyfield
	//	pRecord
	//		Pointer to your keyfield

	DWORD			ReadTableMemoFieldSize(short TableId, short FieldId);
	//Get the size of a memofield
	//Returns the size of the memofield, 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldId
	//		Id of the memofield
	//Use this only after find record command (or query command)

	BOOL			ReadTableMemoField(short TableId, short FieldId, LPVOID pRecord);
	//Read a memofield
	//Returns the size of the memofield, 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldId
	//		Id of the memofield
	//	pRecord
	//		Buffer that will recieve the memofield
	//Use this only after find record command (or query command)

	BOOL			WriteTableMemoField(short TableId, short FieldId, LPVOID pRecord, DWORD MemoSize);
	//Write a memofield
	//Returns the size of the memofield, 0 if error
	//Parameters:
	//	TableId
	//		Id of the table
	//	FieldId
	//		Id of the memofield
	//	pRecord
	//		Buffer that will recieve the memofield
	//	MemoSize
	//		Size of the memofield, 0 = delete memofield
	//Use this only after add or update record command

	WORD			GetUpdateCount(short TableId);
	//Returns the updatecount of a table
	//Parameters:
	//	TableId
	//		Id of the table
	//
	//Use this member in Timer or OnIdle to check if another user has changed 
	//a record in the table
};
